from __future__ import annotations

from typing import Sequence

import jax
import jax.numpy as jnp

from ..typing import *


def distance_similarity(pred_locations: ArrayLike, gt_locations: ArrayLike) -> Array:
    """Compute distance similarity matrix

        pairwise similarity = 1 / distance ^2

    Args:
        pred_locations: [N, d] use -1 to mask out invalid locations
        gt_locations: [K, d] use -1 to mask out invalid locations

    Returns:
        similarity_matrix: [N, k]
    """

    distances_matrix = pred_locations[:, None, :] - gt_locations
    distances_matrix = jnp.square(distances_matrix).sum(axis=-1)

    sm = 1.0 / (distances_matrix + 1e-8)

    # negative locations are invalid
    sm = jnp.where((pred_locations < 0).all(axis=-1)[:, None], 0, sm)
    sm = jnp.where((gt_locations < 0).all(axis=-1), 0, sm)

    return sm


def location_matching(
    pred_locations: ArrayLike, gt_locations: ArrayLike, threshold: float
) -> tuple[Array, Array]:
    """Match predicted location to gt locations

    Args:
      pred_locations:r [N, d]
      gt_locations: [K, d]
      threshold: float. Maximum distance to be matched

    Returns:
      matches: [N], indices of the matches location in gt list
      indicators: [N] bool
    """

    threshold = 1.0 / (threshold * threshold)

    distances_matrix = distance_similarity(pred_locations, gt_locations)

    matches = distances_matrix.argmax(axis=-1, keepdims=True)
    indicators = jnp.take_along_axis(distances_matrix, matches, -1) > threshold

    return matches.squeeze(-1), indicators.squeeze(-1)



from .convnext import ConvNeXt
from .lacss import Lacss
from .lpn import LPN
# from .resnet import ResNet
from .segmentor import Segmentor
from .unet import UNet
from .stack_integrator import StackIntegrator
from .common import *


from __future__ import annotations

from typing import Sequence, Tuple, Callable

import flax.linen as nn
import jax
import jax.numpy as jnp
import optax
from flax.linen import normalization
from flax.linen import activation

from .common import picklable_relu
from ..losses.common import binary_focal_crossentropy
from ..ops import non_max_suppression, location_matching
from ..typing import Array, ArrayLike

EPS = jnp.finfo("float32").eps


class LPN(nn.Module):
    """Location detection head

    Attributes:
        feature_levels: Input feature level, e.g. [0, 1]
        feature_level_scales: The scaling of each feature level. 
        conv_spec: Conv layer specification
        detection_n_cls: number of classes
        nms_threshold: non-max-supression threshold, if performing nms on detected locations.
        pre_nms_topk: max number of detections to be processed regardless of nms, ignored if negative
        max_output: number of detection outputs 

        detection_roi: Parameter for label smoothing for focal loss
        detection_loss_gamma: gamma value for detection focal loss
        detection_loss_delta: delta value for offset regression huber loss
    """

    # network hyperparams
    feature_levels: Sequence[int] = (0, 1, 2)
    feature_level_scales: Sequence[int] = (4, 8, 16)
    conv_spec: Sequence[int] = (192, 192, 192, 192)
    normalization: Callable[[None], nn.Module]=normalization.GroupNorm
    activation: Callable[[Array], Array] = picklable_relu

    detection_n_cls: int = 1

    # loss hyperparams
    detection_roi: float = 8.0
    detection_loss_gamma: float = 2.0
    detection_loss_delta: float = 1.0

    # detection hyperparams
    nms_threshold: float = 8.0
    pre_nms_topk: int = -1
    max_output: int = 512
    min_score: float = 0.2
    max_proposal_offset: float = 12.0
    z_scale:float = 5.0

    def _best_match(self, gt_locations, pred_locations):
        """replacing gt_locations with pred_locations if the close enough
        1. Each pred_location is matched to the closest gt_location
        2. For each gt_location, pick the matched pred_location with highest score
        3. if the picked pred_location is within threshold distance, replace the gt_location with the pred_location
        """

        threshold = self.max_proposal_offset

        n_gt_locs = gt_locations.shape[0]
        n_pred_locs = pred_locations.shape[0]

        matched_id, indicators = location_matching(
            pred_locations * jnp.asarray([self.z_scale, 1, 1]), 
            gt_locations * jnp.asarray([self.z_scale, 1, 1]), 
            threshold,
        )
        matched_id = jnp.where(indicators, matched_id, -1)

        matching_matrix = (
            matched_id[None, :] == jnp.arange(n_gt_locs)[:, None]
        )  # true at matched gt(row)/pred(col), at most one true per col
        last_col = jnp.ones([n_gt_locs, 1], dtype=bool)
        matching_matrix = jnp.concatenate(
            [matching_matrix, last_col], axis=-1
        )  # last col is true

        # first true of every row
        matched_loc_ids = jnp.argmax(matching_matrix, axis=-1)

        training_locations = jnp.where(
            matched_loc_ids[:, None] == n_pred_locs,  # true: failed match
            gt_locations,
            pred_locations[
                matched_loc_ids, :
            ],  # out-of-bound error silently dropped in jax
        )

        return training_locations

    def _to_targets(
        self,
        locations: ArrayLike,
        clses: ArrayLike,
        feature_shape: tuple[int, int],
        scale: float,
    ) -> tuple[Array, Array]:
        """Generate labels as LPN regression targets

        Returns:
            is_target: [H, W, n_cls] int32
            regression_target: [H, W, 2] float tensor
        """
        depth, height, width = feature_shape

        threshold_sq = self.detection_roi ** 2

        flat_locations = jnp.expand_dims(
            locations * jnp.asarray([self.z_scale, 1, 1]),
            (1, 2, 3),
        ) #[N, 1, 1, 1, 3]

        mesh = (
            jnp.moveaxis(jnp.mgrid[:depth, :height, :width] + 0.5, 0, -1) 
            * jnp.asarray([self.z_scale, scale, scale])
        )  # [D, H, W, 3]
        distances = flat_locations - mesh  # [N, D, H, W, 3]
        distances_sq = (distances * distances).sum(axis=-1)  # [N, D, H, W]

        # masking off invalid
        distances_sq = jnp.where(
            (flat_locations >= 0).all(axis=-1), distances_sq, float("inf")
        )

        indices = jnp.argmin(
            distances_sq, axis=0, keepdims=True
        )  # [1, D, H, W] map to nearest gt-loc
        cls_map = clses[
            indices.squeeze(0)
        ]  # [D, H, W] map to the cls of the nearest gt-loc

        best_distances = jnp.take_along_axis(distances_sq, indices, 0).squeeze(0)
        is_target = (
            best_distances < threshold_sq
        )  # [D, H, W] indicate whether the pixel is close to a gt-loc
        cls_target = jnp.where(is_target, cls_map, self.detection_n_cls)  # [D, H, W]

        indices = indices[...,None].repeat(3, axis=-1)  # [1, D, H, W, 3]
        regression_target = jnp.take_along_axis(distances, indices, 0)
        regression_target = regression_target.squeeze(0) # [D, H, W, 3]
        regression_target = regression_target / jnp.asarray([self.z_scale, scale, scale]) # normalize

        return cls_target, regression_target

    def _compute_losses(self, network_outputs, gt_locations, gt_cls):
        cls_loss, regr_loss = [], []
        n_targets = 0

        for block_out in network_outputs.values():
            depth, height, width, _ = block_out["cls_logits"].shape

            if gt_cls is None:
                gt_cls = jnp.zeros(gt_locations.shape[0], dtype=int)

            cls_target, regression_target = self._to_targets(
                gt_locations,
                gt_cls,
                (depth, height, width),
                block_out["scale"],
            )

            self.sow("intermediates", "cls_target", cls_target)
            self.sow("intermediates", "regression_target", regression_target)

            cls_loss.append(
                binary_focal_crossentropy(
                    jax.nn.softmax(block_out["cls_logits"]),
                    jax.nn.one_hot(cls_target, self.detection_n_cls+1),
                    gamma=self.detection_loss_gamma,
                ).mean(axis=-1).reshape(-1)
            )

            regr_loss.append(
                jnp.where(
                    cls_target != self.detection_n_cls,
                    optax.huber_loss(
                        block_out["regressions"],
                        regression_target,
                        delta=self.detection_loss_delta,
                    ).mean(axis=-1),
                    0,
                ).reshape(-1)
            )

            n_targets += jnp.count_nonzero(cls_target != self.detection_n_cls)

        cls_loss = jnp.concatenate(cls_loss).mean()
        regr_loss = jnp.concatenate(regr_loss).sum() / (n_targets + 1e-8)

        return dict(
            lpn_localization_loss=regr_loss,
            lpn_detection_loss=cls_loss,
        )

    def _generate_predictions(self, network_outputs):
        """
        Produce a list of proposal locations based on predication map, remove redundency with non_max_suppression
        """
        distance_threshold = self.nms_threshold
        output_size = self.max_output
        topk = self.pre_nms_topk
        score_threshold = self.min_score

        network_outputs = jax.lax.stop_gradient(network_outputs)

        # preprocess
        scores, locations, clss = [], [], []
        for block_out in network_outputs.values():
            logits = jax.nn.softmax(block_out["cls_logits"])
            scores_ = logits[..., :-1].sum(axis=-1) # [D, H, W]
            clss_ = logits[..., :-1].argmax(axis=-1)

            depth, height, width = scores_.shape
            loc_ = jnp.moveaxis(jnp.mgrid[:depth, :height, :width] + 0.5, 0, -1)  # [D, H, W, 3]
            loc_ = loc_ + block_out["regressions"]

            # set invalid location score to -1
            max_values = jnp.asarray([depth, height, width])
            is_valid = (loc_ > 0.0).all(axis=-1) & (loc_ < max_values).all(axis=-1)
            scores_ = jnp.where(is_valid, scores_, -1)

            loc_ = loc_ * jnp.asarray([self.z_scale, block_out["scale"], block_out["scale"]]) # back to physical scale

            scores.append(scores_.reshape(-1))
            locations.append(loc_.reshape(-1, 3))
            clss.append(clss_.reshape(-1))

        scores, locations, clss = map(jnp.concatenate, (scores, locations, clss))

        assert scores.shape[0] == locations.shape[0]
        assert scores.shape[0] == clss.shape[0]

        # sort and nms
        if topk <= 0 or topk > scores.size:
            topk = scores.size

        scores, selections = jax.lax.top_k(scores, topk)
        locations = locations[selections]
        clss = clss[selections]

        threshold = 1 / distance_threshold / distance_threshold
        scores, locations, indices = non_max_suppression(
            scores,
            locations,
            output_size,
            threshold,
            score_threshold,
            return_selection=True,
        )

        idx_of_selected = jnp.argwhere(
            indices, size=output_size, fill_value=-1
        ).squeeze(-1)
        clss = jnp.where(idx_of_selected >= 0, clss[idx_of_selected], -1)

        locations = locations / jnp.asarray([self.z_scale, 1, 1]) # reverse z scaling

        return dict(
            scores=scores,
            locations=locations,
            classes=clss,
        )

    @nn.compact
    def _block(self, feature: ArrayLike) -> Tuple[ArrayLike, ArrayLike, ArrayLike]:
        conv_spec = self.conv_spec

        x = feature

        for n_ch in conv_spec:
            x = nn.Conv(n_ch, (3, 3))(x)
            x = self.normalization()(x)
            x = self.activation(x)

        self.sow("intermediates", "lpn_features", x)

        cls_logits = nn.Conv(self.detection_n_cls + 1, (1, 1))(x)
        regressions = nn.Conv(3, (1, 1))(x)

        # no z-localization for 2d input
        if x.shape[0] == 1:
            regressions = regressions.at[..., 0].set(0.)

        return dict(
            regressions=regressions,
            cls_logits=cls_logits,
        )

    def __call__(
        self,
        inputs: Sequence[ArrayLike],
        gt_locations: ArrayLike | None = None,
        gt_cls: ArrayLike | None = None,
        *,
        training: bool = False,
        **kwargs,
    ) -> dict:
        """
        Args:
            inputs: a list of features: Sequence[Array...]
            gt_locations: [N, 3]
            gt_cls: class ids. [N]

        Returns:
            A dict of model output
        """
        network_outputs = {}

        for lvl, scale in zip(self.feature_levels, self.feature_level_scales):
            feature = inputs[lvl]
            network_outputs[str(lvl)] = self._block(feature)
            network_outputs[str(lvl)]["scale"] = scale

        # prediction
        predictions = self._generate_predictions(network_outputs)

        # losses
        if training:
            losses = self._compute_losses(network_outputs, gt_locations, gt_cls)
            network_outputs["training_locations"] = self._best_match(
                gt_locations, 
                predictions["locations"],
            )

        else:
            losses = {}

        # format outputs
        return dict(
            detector=network_outputs,
            predictions=predictions,
            losses=losses,
        )



from __future__ import annotations

from dataclasses import field

import flax.linen as nn
import jax.numpy as jnp

from ..ops import *
from ..typing import *
from ..utils import deep_update
from .convnext import ConvNeXt
from .lpn import LPN
from .segmentor import Segmentor
from .stack_integrator import StackIntegrator

class Lacss(nn.Module):
    """Main class for LACSS model

    Attributes:
        stem: Prepprocessing module
        backbone: by default a ConvNeXt CNN
        integrator: module for integrating backbone output
        detector: detection head to predict cell locations
        segmentor: The segmentation head
    """

    stem: nn.Module | None = None
    backbone: nn.Module = field(default_factory=ConvNeXt)
    integrator: nn.Module | None = field(default_factory=StackIntegrator)
    detector: nn.Module = field(default_factory=LPN)
    segmentor: nn.Module | None = field(default_factory=Segmentor)

    def __call__(
        self,
        image: ArrayLike,
        gt_locations: ArrayLike | None = None,
        gt_cls: ArrayLike | None = None,
        mask: ArrayLike | None = None,
        *,
        training: bool | None = None,
    ) -> dict:
        """
        Args:
            image: [H, W, C] or [D, H, W, C]
            gt_locations: [M, 2/3] locations of the instances
            gt_cls: [M] instance classes for multi-class prediction
            mask: broadcastable boolean mask indicating valid region of the input image
        Returns:
            a dict of model outputs

        """
        outputs = {}

        x = image

        # if inputs are 2D
        if x.ndim == 3:
            x = x[None, ...]
        assert x.ndim == 4
        
        if mask is not None:
            if mask.ndim == 2:
                mask = mask[None, ...]
            assert mask.ndim==3

        if gt_locations is not None:
            if gt_locations.shape[-1] == 2:
                gt_locations = jnp.c_[jnp.zeros_like(gt_locations[:, :1]) + .5, gt_locations]
        
            if gt_cls is None:
                gt_cls = jnp.zeros(gt_locations.shape[0], dtype=int)

            assert gt_locations.ndim == 2 and gt_locations.shape[-1] == 3

        if self.stem is not None:
            x = self.stem(x, mas=mask, training=training)

        x = self.backbone(x, mask=mask, training=training)
        self.sow("intermediates", "encoder_features", x)

        if self.integrator is not None:
            x = self.integrator(x, mask=mask, training=training)
            self.sow("intermediates", "decoder_features", x)

        detector_out = self.detector(x, gt_locations, gt_cls, mask=mask, training=training)
        outputs = deep_update(outputs, detector_out)

        if self.segmentor is not None:
            if training:
                seg_locs= outputs["detector"]["training_locations"]
                seg_cls = gt_cls
            elif gt_locations is not None: # not training, but locations of cells are known
                seg_locs, seg_cls = gt_locations, gt_cls
            else:
                seg_locs = outputs["predictions"]["locations"]
                seg_cls = outputs["predictions"]["classes"]

            self.sow("intermediates", "segmentation_locations", seg_locs)
            self.sow("intermediates", "segmentation_cls", seg_cls)

            segmentor_out = self.segmentor(x, seg_locs, seg_cls, mask=mask, training=training)

            outputs = deep_update(outputs, segmentor_out)

        return outputs

    def get_config(self):
        from dataclasses import asdict
        from ..utils import remove_dictkey

        cfg = asdict(self)

        remove_dictkey(cfg, "parent")

        return cfg

    @classmethod
    def from_config(cls, config):
        from collections import defaultdict

        if isinstance(config, str):
            return cls.get_preconfigued(config)

        else:
            config_ = defaultdict(lambda: {})
            config_.update(config)
            return Lacss(
                backbone=ConvNeXt(**config_["backbone"]),
                integrator=StackIntegrator(**config_["integrator"]),
                detector=LPN(**config_["detector"]),
                segmentor=Segmentor(**config_["segmentor"]),
            )


    @classmethod
    def get_default_model(cls, patch_size=4):
        if not patch_size in (1, 2, 4):
            raise ValueError("patch_size must be 1, 2 or 4")
        return cls(
            backbone=ConvNeXt.get_model_small(patch_size=patch_size),
            integrator=StackIntegrator(dim_out=384),
            detector=LPN(
                conv_spec=(256, 256, 256, 256),
                feature_levels=(0, 1, 2),
                feature_level_scales=(patch_size, patch_size * 2, patch_size * 4),
            ),
            segmentor=Segmentor(
                conv_spec=((256, 256, 256), (64,)),
                feature_level=0,
                feature_scale=patch_size,
            ),
        )

    @classmethod
    def get_small_model(cls, patch_size=4):
        return cls(
            backbone=ConvNeXt.get_model_tiny(patch_size=patch_size),
            integrator=StackIntegrator(dim_out=256),
            detector=LPN(
                conv_spec=(192, 192, 192, 192),
                feature_levels=(0, 1, 2),
                feature_level_scales=(patch_size, patch_size * 2, patch_size * 4),
            ),
            segmentor=Segmentor(
                conv_spec=((192, 192, 192), (48,)),
                feature_level=0,
                feature_scale=patch_size,
            ),
        )

    @classmethod
    def get_large_model(cls, patch_size=4):
        return cls(
            backbone=ConvNeXt.get_model_base(patch_size=patch_size),
            integrator=StackIntegrator(dim_out=512),
            detector=LPN(
                conv_spec=(384, 384, 384, 384),
                feature_levels=(0, 1, 2),
                feature_level_scales=(patch_size, patch_size * 2, patch_size * 4),
            ),
            segmentor=Segmentor(
                conv_spec=((384, 384, 384), (128,)),
                feature_level=0,
                feature_scale=patch_size,
            ),
        )

    @classmethod
    def get_preconfigued(cls, config: str, *, patch_size: int = 4):
        if config == "default" or config == "base":
            return cls.get_default_model(patch_size)
        elif config == "small":
            return cls.get_small_model(patch_size)
        elif config == "large":
            return cls.get_large_model(patch_size)
        else:
            raise ValueError(f"Unkown model config {config}")



from __future__ import annotations

import pathlib
import typing

import optax
from jax import Array
from jax.typing import ArrayLike

Optimizer = optax.GradientTransformation

PathLike = typing.Union[str, pathlib.Path]

DataDict = typing.Mapping[str, ArrayLike]

Patches = DataDict




from __future__ import annotations

import math
from typing import Sequence, Callable

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax.linen import normalization
from flax.linen import activation

from lacss.ops import sub_pixel_samples

from ..typing import ArrayLike, Array
from .common import SpatialAttention, picklable_relu


class Segmentor(nn.Module):
    """LACSS segmentation head.

    Attributes:
        feature_level: the feature level used for segmentation
        feature_scale: the spatail scale of the feature level
        conv_spec: conv_block definition, e.g. ((384,384,384), (64,))
        instance_crop_size: Crop size for segmentation.
        with_attention: Whether use spatial attention layer.
        encoder_dims: Dim of the learned position encoder.
    """
    feature_level: int = 0
    feature_scale: int = 4
    conv_spec: tuple[Sequence[int], Sequence[int]] = (
        (192, 192, 192),
        (48,),
    )
    normalization: Callable[[None], nn.Module]=normalization.GroupNorm
    activation: Callable[[Array], Array] = picklable_relu

    n_cls: int = 1
    instance_crop_size: int = 96
    use_attention: bool = False
    encoder_dims: Sequence[int] = (8, 8, 4)
    max_embedding_length:int = 16

    @property
    def patch_size(self):
        return self.instance_crop_size // self.feature_scale

    def _gather_patches(self, feature, locs, patch_size=None):
        patch_size = patch_size or self.patch_size
        locs = locs.astype(int) # so we can use loc for indexing

        yy, xx = (
            jnp.mgrid[:patch_size, :patch_size] 
            - patch_size // 2
            + locs[:, :, None, None]
        ).swapaxes(0, 1) # [2, N, PS, PS]

        # padding to avoid out-of-bound
        padding_size = patch_size // 2 + 1
        paddings = [
            [padding_size, padding_size],
            [padding_size, padding_size],
            [0, 0],
        ]
        padded_feature = jnp.pad(feature, paddings)

        patches = padded_feature[yy + padding_size, xx + padding_size, :]

        return patches, yy, xx

    def _pos_encoder(self, feature, locations):
        # derive pos embedding (YX) from feature values at cell location
        # the rationale is that the feature value already has some info regarding
        # cell size/shape
        patch_size = self.patch_size
        depth = feature.shape[0]

        sampling_locs = jnp.floor(locations) - jnp.asarray([0, 0.5, 0.5]) # center of the patch
        x = sub_pixel_samples(feature, sampling_locs)
        assert x.shape == (locations.shape[0], feature.shape[-1])

        for _ in range(2):
            x = nn.Dense(x.shape[-1])(x)
            x = nn.relu(x)

        dim_out = math.prod(self.encoder_dims[-3:])
        xa = nn.Dense(dim_out)(x)
        xa = xa.reshape((-1,) + self.encoder_dims[-3:])

        xa = jax.image.resize(
            xa,
            (
                x.shape[0],
                patch_size,
                patch_size,
                self.encoder_dims[-1],
            ),
            "linear",
        )

        encodings = nn.Dense(self.conv_spec[1][0], use_bias=False)(xa)

        if len(self.encoder_dims) == 3:
            z_encodings = self._z_encodings(depth, locations[:, 0]) # [D, N, c]

        else:
            z_dim = self.encoder_dims[0] * self.encoder_dims[-1]
            xb = nn.Dense(z_dim)(x)
            xb = xb.reshape(-1, self.encoder_dims[0], self.encoder_dims[-1])
            xb = jax.image.resize(
                xb,
                (
                    x.shape[0],
                    self.max_embedding_length,
                    self.encoder_dims[-1],
                ),
                "linear",
            )
            embeddings = nn.Dense(self.conv_spec[1][0], use_bias=False)(xb) # [N, L, c]

            z_idx = self.max_embedding_length // 2 - locations[:, 0].astype(int) # [N]
            z_idx = jnp.clip(
                z_idx[:, None] + jnp.arange(depth), 
                0, self.max_embedding_length -1
            ) # [N, D]

            z_encodings = jax.vmap(lambda b, z: b[z])(embeddings, z_idx) # [N, D, c]
            z_encodings = z_encodings.swapaxes(0, 1) # [D, N, c]
                        
        encodings = encodings + z_encodings[:, :, None, None, :] #[D, N, ps, ps, c]

        return encodings #[D, N, ps, ps, c]

    def _z_encodings(self, depth, z_locs):
        # z_pos_embedding is a simple learned matrix
        max_embedding_length = self.max_embedding_length
        target_dim = self.conv_spec[1][0]

        pos_embedding = self.param(
            "z_embedding",
            nn.initializers.normal(stddev=0.02),
            (max_embedding_length, target_dim),
        )

        z_idx = max_embedding_length // 2 - z_locs.astype(int)
        z_idx = jnp.clip(
            z_idx[:, None] + jnp.arange(depth),
            0, self.max_embedding_length -1,
        ) # [N, D]

        pos_embedding = jax.vmap(lambda z: pos_embedding[z])(z_idx) # [N, D, n_ch]
        pos_embedding = pos_embedding.swapaxes(0, 1) # [D, N, n_ch]

        return pos_embedding #[D, N, n_ch]

    @nn.compact
    def __call__(
        self,
        features: list[ArrayLike],
        locations: ArrayLike,
        classes: ArrayLike | None = None,
        mask: ArrayLike | None = None,
        *,
        training=False,
    ) -> dict:
        """
        Args:
            features: list[array: [D, H, W, C]] multiscale features from the backbone.
            locations: [N, 3]
            classes: [N] optional

        Returns:
            A nested dictionary of values representing segmentation outputs.
              * segmentor/logits: all segment logits
              * predictions/segmentations: logits of designated class [N, D, ps, ps]
              * predictions/segmentation_y0_coord: y coord of patch top-left corner [N]
              * predictions/segmentation_x0_coord: x coord of patch top-left corner [N]
              * predictions/segmentation_is_valid: mask for valid patches [N]
        """
        assert features[self.feature_level].ndim == 4
        assert locations.ndim == 2
        assert locations.shape[-1] == 3

        crop_size = self.instance_crop_size
        locations = locations / jnp.asarray([1, self.feature_scale, self.feature_scale])

        # feature convs
        x = features[self.feature_level]
        D, H, W, C = x.shape

        for n_ch in self.conv_spec[0]:
            x = nn.Conv(n_ch, (3, 3))(x)
            x = self.normalization()(x)
            x = self.activation(x)

        x = nn.Conv(self.conv_spec[1][0], (3, 3))(x)

        self.sow("intermediates", "segmentor_image_features", x) # [D, H, W, C]

        # pos encoder
        feature = features[self.feature_level]
        pos_encodings = self._pos_encoder(feature, locations) # [D, N, ps, ps, c]

        self.sow("intermediates", "segmentor_pos_embed", pos_encodings)

        # mixing features with pos_encodings
        patches, ys, xs = jax.vmap(
            self._gather_patches,
            in_axes=(0, None),
            out_axes=(0, None, None),
        )(x, locations[:, 1:]) # [D, N, ps, ps, c]

        patches = jax.nn.relu(patches + pos_encodings) 

        if self.use_attention:
            patches = SpatialAttention()(patches)

        # patche convs
        for n_ch in self.conv_spec[1][1:]:
            patches = nn.Conv(n_ch, (3, 3))(patches)
            patches = self.activation(patches)

        self.sow("intermediates", "segmentor_patch_features", patches)

        # outputs
        if self.feature_scale == 1:
            logits = nn.Conv(self.n_cls, (3, 3))(patches)
        else:
            logits = nn.ConvTranspose(self.n_cls, (2, 2), strides=(2, 2))(patches)
            if self.feature_scale != 2:
                logits = jax.image.resize(
                    logits,
                    patches.shape[:2] + (crop_size, crop_size, self.n_cls),
                    "linear",
                )

        outputs = jnp.take_along_axis(
            logits,
            classes.reshape(1, -1, 1, 1, 1),
            axis=-1,
        ) # [D, N, ps, ps, 1]
        outputs = outputs.swapaxes(0, 1).squeeze(-1) #[N, D, ps, ps]

        # clear invalid locations, needed for the semi-supervied loss functions
        instance_mask = (locations >= 0).all(axis=-1)
        y0 = jnp.where(instance_mask, ys[:, 0, 0] * self.feature_scale, -1)
        x0 = jnp.where(instance_mask, xs[:, 0, 0] * self.feature_scale, -1)
        yc, xc = jnp.mgrid[:self.instance_crop_size, :self.instance_crop_size]
        yc = yc + y0[:, None, None] # [N, ps, ps]
        xc = xc + x0[:, None, None]
        valid_pixels = (xc >= 0) & (yc >=0) & (xc < W * self.feature_scale) & (yc < H * self.feature_scale)
        if mask is not None:
            valid_pixels &= mask.any(axis=0)[yc, xc] # [N, ps, ps]
            valid_pixels = jnp.expand_dims(valid_pixels, 1) & mask.any(axis=(1,2), keepdims=True) #[N, D, ps, ps]
        else:
            valid_pixels = jnp.expand_dims(valid_pixels, 1)
        valid_pixels &= jnp.expand_dims(instance_mask, (1,2,3))
        outputs = jnp.where(
            valid_pixels,
            outputs,
            -1e8,
        )

        return dict(
            segmentor=dict(
                logits=logits,
            ),
            predictions=dict(
                segmentations=outputs, #[N, D, ps, ps]
                segmentation_y0_coord=y0,
                segmentation_x0_coord=x0,
                segmentation_is_valid=instance_mask,
            ),
        )



from __future__ import annotations

from typing import Sequence, Callable

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax.linen import normalization
from flax.linen import activation

from ..typing import ArrayLike, Array
from .common import FPN, FFN

Initializer = Callable[[jnp.ndarray, Sequence[int], jnp.dtype], jnp.ndarray]

_MAX_EMBEDDING_LENGTH = 32
_EMBEDDING_SCALE = 64

class _StackTransformer(nn.Module):
    n_layers: int = 2
    n_heads: int = 4
    dropout: float = 0.
    ff_dropout: float = 0.6
    deterministic: bool|None = None
    learned_embedding: bool = False

    def _get_pos_embeddings(self, feature):
        n_ch = feature.shape[-1]

        if self.learned_embedding:
            pos_embedding = self.param(
                'pos_embedding', 
                nn.initializers.normal(stddev=0.02), 
                (_MAX_EMBEDDING_LENGTH, n_ch),
                feature.dtype,
            )
        else:
            pos_embedding = np.zeros((_MAX_EMBEDDING_LENGTH, n_ch))
            pos = np.arange(_MAX_EMBEDDING_LENGTH)
            pos = pos[:, None] * np.exp(- np.log(_EMBEDDING_SCALE) / n_ch * np.arange(0, n_ch, 2))
            pos_embedding[:, 0::2] = np.sin(pos)
            pos_embedding[:, 1::2] = np.cos(pos)
        
        return jnp.asarray(pos_embedding)

    def _attention(self, feature, pos_embedding, mask, deterministic):
        shortcut = feature

        if mask is not None:
            mask = mask.any(axis=(1,2))

        feature = nn.LayerNorm()(feature)
        feature = nn.MultiHeadDotProductAttention(
            self.n_heads,
            dropout_rate=self.dropout,
            broadcast_dropout=False,
        )(
            inputs_q = feature + pos_embedding,
            inputs_k = feature + pos_embedding,
            inputs_v = feature,
            deterministic=deterministic,
            sow_weights=True,
            mask=mask,
        )
        feature = shortcut + feature

        return feature

    @nn.compact
    def __call__(self, feature:ArrayLike, mask, deterministic:bool|None=None) -> Array:
        assert feature.ndim == 4, f"Wrong input dimension to StackIntegrator: {feature.shape}"

        depth, height, width, n_ch = feature.shape

        if deterministic is None:
            deterministic = self.deterministic

        # bypass attention for 2d
        if feature.shape[0] == 1:
            return feature

        pos_embedding = self._get_pos_embeddings(feature)
        pos_embedding = pos_embedding[:depth]
        assert pos_embedding.shape == (depth, n_ch)

        feature = jnp.moveaxis(feature, 0, 2) # we are integrating along z axis
        for _ in range(self.n_layers):
            feature = self._attention(feature, pos_embedding, mask, deterministic=deterministic)
            feature = FFN(dropout_rate=self.ff_dropout)(feature, deterministic=deterministic)        
        feature = jnp.moveaxis(feature, 2, 0) # move the z axis back

        return feature


class StackIntegrator(nn.Module):
    """ Feature integration for image stacks.

    This applies a simple transformer on the stack of features at the lowest resolution to integrate information
    from all inputs images. This computation is very light-weight, becasue the integration happens only for 
    features at the same location -- the y-x axis are treated as batch dimension. After transformer, the features
    at different resolutions are mixed using a standard FPN.

    Attributes:
      n_layers: number transformer layers
      n_heads: number of heads

      dropout: transformer attention weight dropout rate
      ff_dropout: the dropout rate of the FFN part of the transformer

      dim_out: the final output feature dim (after FPN)
    """
    n_layers: int = 2
    n_heads: int = 4
    dim_out: int = 256
    dropout: float = 0.
    ff_dropout: float = 0.2

    @nn.compact
    def __call__(self, features:Sequence[ArrayLike], mask, *, training:bool=False) -> Sequence[Array]:
        feature_last = features[-1]
        feature_last = _StackTransformer(
            self.n_layers,
            self.n_heads,
            self.dropout,
            self.ff_dropout,
        )(
            feature_last, mask, deterministic=not training,
        )

        features = list(features[:-1]) + [feature_last]

        return FPN(self.dim_out)(features, training=training)



from __future__ import annotations

from random import random
import numpy as np

def _uniform_rand(min_v, max_v):
    return random() * (max_v - min_v) + min_v

def _init_boolean_mask(inputs):
    if "centroids" in inputs:
        return np.ones([len(inputs["centroids"])], dtype=bool)
    elif "bboxes" in inputs:
        return np.ones([len(inputs["bboxes"])], dtype=bool)
    else:
        return None

def _value_pair(x):
    try:
        a, b = x
    except:
        a = x
        b = x

    return a, b

def _image_size(img):
    H = img.shape[-3]
    W = img.shape[-2]

    return (H, W)

#FIXME try a np implementation
def _crop_and_resize(masks, boxes, target_shape):
    import jax
    from lacss.ops import sub_pixel_crop_and_resize
    return np.asarray(jax.vmap(
        sub_pixel_crop_and_resize,
        in_axes=(0, 0, None),
    )(masks, boxes, target_shape))

def resize(inputs: dict, *, target_size: tuple[int, int], p: float = 1.0) -> dict:
    """Resize image and labels

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target size
        p: probability of applying transformation
    """

    def _resize(inputs):
        from skimage.transform import resize
        H, W = _image_size(inputs["image"])

        target_y, target_x = target_size

        scaling_y = target_y / H
        scaling_x = target_x / W

        inputs["image"] = resize(
            inputs["image"],
            [target_y, target_x],
            antialias=True,
        )

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"] / inputs["image_mask"].max() 
            image_mask = resize(image_mask, [target_y, target_x])
            image_mask = (image_mask >= 0.5).astype(inputs["image_mask"].dtype)
            inputs["image_mask"] = image_mask

        if "centroids" in inputs:
            inputs["centroids"] *= np.asarray([scaling_y, scaling_x])

        if "bboxes" in inputs:
            inputs["bboxes"] *= np.asarray([scaling_y, scaling_x, scaling_y, scaling_x])
            
        return inputs

    if p < 1.0 and random() >= p:
        return inputs

    else:
        return _resize(inputs)


def random_resize(
    inputs: dict,
    *,
    scaling: float | tuple[float, float],
    keep_aspect_ratio: bool = False,
    p: float = 1.0,
) -> dict:
    """Resize image by a random amount

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids

    Keyword Args:
        scaling: range of scale, e.g, [0.8, 1.5].
        keep_aspect_ratio: Whether to scale x/y the same amount.
        p: probability of applying transformation

    """

    if p < 1.0 and random() >= p:
        return inputs

    else:
        min_scale, max_scale = scaling
        H, W = _image_size(inputs["image"])

        if keep_aspect_ratio:
            scaling_x = scaling_y = _uniform_rand(min_scale, max_scale)
        else:
            scaling_y = _uniform_rand(min_scale, max_scale)
            scaling_x = _uniform_rand(min_scale, max_scale)

        target_h = scaling_y * H
        target_w = scaling_x * W

        return resize(inputs, target_size=(target_h, target_w))


def crop_to_roi(
    inputs: dict,
    *,
    roi: tuple[int, int, int, int],
    area_ratio_threshold: float = 1.0,
    p: float = 1.0,
) -> dict:
    """Crop image to bounding-box ROI

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        roi: Rectangle roi in yxyx format
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation

    """

    def _crop_to_roi(inputs):
        y0, x0, y1, x1 = roi
        roi_h = int(y1 - y0)
        roi_w = int(x1 - x0)

        inputs["image"] = inputs["image"][y0:y1, x0:x1, :]

        if "image_mask" in inputs:
            inputs["image_mask"] = inputs["image_mask"][y0:y1, x0:x1]

        boolean_mask = _init_boolean_mask(inputs)

        if boolean_mask is None:
            return inputs

        if "centroids" in inputs:
            ctrds = inputs["centroids"] - [y0, x0]
            boolean_mask &= (ctrds > 0).all(axis=-1) & (ctrds < [roi_h, roi_w]).all(axis=-1)

        if "bboxes" in inputs:
            bboxes = inputs["bboxes"] - [y0, x0, y0, x0]
            areas = (bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])

            bboxes = np.clip(bboxes, 0, [roi_h, roi_w, roi_h, roi_w])
            new_areas = (bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])

            boolean_mask &= (new_areas / areas) >= area_ratio_threshold

        if "centroids" in inputs:
            inputs["centroids"] = ctrds[boolean_mask]

        if "bboxes" in inputs:
            orig_bboxes = inputs["bboxes"]
            inputs["bboxes"] = bboxes[boolean_mask]

            if "masks" in inputs:
                target_shape = inputs["masks"].shape[1:3]

                orig_bboxes = orig_bboxes[boolean_mask]
                orig_y0 = orig_bboxes[:, 0] - y0
                orig_x0 = orig_bboxes[:, 1] - x0
                orig_h = orig_bboxes[:, 2] - orig_bboxes[:, 0]
                orig_w = orig_bboxes[:, 3] - orig_bboxes[:, 1]

                new_bbox = inputs["bboxes"] - [orig_y0, orig_x0, orig_y0, orig_x0]
                new_bbox = new_bbox / [orig_h, orig_w, orig_h, orig_w]
                new_bbox = new_bbox * (target_shape + target_shape)

                inputs["masks"] = _crop_and_resize(
                    inputs["masks"][boolean_mask],
                    new_bbox,
                    target_shape,
                )

        return inputs

    if p < 1.0 and random() >= p:
        return inputs

    else:
        return _crop_to_roi(inputs)


def random_crop(
    inputs: dict,
    *,
    target_size: tuple[int, int],
    area_ratio_threshold: float = 1.0,
    p: float = 1.0,
) -> dict:
    """Random crop to a set target size

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        target_size: the target size
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation
    """

    H, W = _image_size(inputs["image"])

    target_h, target_w = _value_pair(target_size)

    if p < 1.0 and random() >= p:
        return inputs

    else:
        y0 = int(random() * (H - target_h))
        x0 = int(random() * (W - target_w))

        return crop_to_roi(
            inputs,
            roi=(y0, x0, y0 + target_h, x0 + target_w),
            area_ratio_threshold=area_ratio_threshold,
        )


def pad(
    inputs: dict,
    *,
    paddings: int | tuple[int, int],
    constant_values: float = 0,
    p: float = 1.0,
) -> dict:
    """Pad image and labels.

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        paddings: either a tuple or a single value. If latter, use the same padding for both x and y axis
        constant_values: the value to fill the padded area
        p: probability of applying transformation
    """

    def _pad(inputs):
        padding_y, padding_x = _value_pair(paddings)
        padding_y0, padding_y1 = _value_pair(padding_y)
        padding_x0, padding_x1 = _value_pair(padding_x)

        inputs["image"] = np.pad(
            inputs["image"],
            [[padding_y0, padding_y1], [padding_x0, padding_x1], [0, 0]],
            constant_values=constant_values,
        )

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"]
            if len(image_mask.shape) == 2:
                padding_values = [[padding_y0, padding_y1], [padding_x0, padding_x1]]
            else:
                padding_values = [
                    [0, 0],
                    [padding_y0, padding_y1],
                    [padding_x0, padding_x1],
                ]
            inputs["image_mask"] = np.pad(
                inputs["image_mask"],
                padding_values,
                constant_values=constant_values,
            )

        if "centroids" in inputs:
            inputs["centroids"] += np.asarray([padding_y0, padding_x0])

        if "bboxes" in inputs:
            inputs["bboxes"] += np.asarray(
                [padding_y0, padding_x0, padding_y0, padding_x0], 
            )

        return inputs

    if p < 1.0 and random() >= p:
        return inputs

    else:
        return _pad(inputs)


def pad_to_size(
    inputs: dict,
    *,
    target_size: tuple[int, int],
    constant_values: float = 0,
    p: float = 1.0,
) -> dict:
    """Pad image and labels to a target size. Padding is applied so that the orginal scene is centered in the output.

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target image size
        constant_values: the value to fill the padded area
        p: probability of applying transformation
    """

    if p < 1.0 and random() >= p:
        return inputs

    H, W = _image_size(inputs["image"])
    target_h, target_w = _value_pair(target_size)

    padding_h = target_h - H
    padding_h = [padding_h // 2, padding_h - padding_h // 2]
    padding_w = target_w - W
    padding_w = [padding_w // 2, padding_w - padding_w // 2]

    return pad(inputs, paddings=[padding_h, padding_w], constant_values=constant_values)


def random_crop_or_pad(
    inputs: dict,
    *,
    target_size: tuple[int, int],
    constant_values: float = 0,
    area_ratio_threshold: float = 1.0,
    p: float = 1.0,
) -> dict:
    """Random crop or pad image to specified target_size.

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        target_size: target size
        constant_values: the value to fill the padded area
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation
    """

    if p < 1.0 and random() >= p:
        return inputs

    else:
        H, W = _image_size(inputs["image"])

        target_h, target_w = _value_pair(target_size)

        h = max(H, target_h)
        w = max(W, target_w)

        inputs = pad_to_size(
            inputs, target_size=(h, w), constant_values=constant_values
        )

        inputs = random_crop(
            inputs, target_size=target_size, area_ratio_threshold=area_ratio_threshold
        )

        return inputs


def flip_left_right(inputs: dict, *, p: float = 1.0) -> dict:
    """Flip image left-right

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        p: probability of applying transformation

    """

    def _flip_left_right(inputs):
        H, W = _image_size(inputs["image"])

        inputs["image"] = inputs["image"][:, ::-1, :]

        if "image_mask" in inputs:
            inputs["image_mask"] = inputs["image_mask"][:, ::-1]

        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * [1, -1] + [0, W]

        #FIXME reformat bboxes so no negative length
        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * [1, -1, 1, -1] + [0, W, 0, W]

        # if "masks" in inputs:
        #     inputs["masks"] = inputs["masks"][:, :, ::-1, :]

        return inputs

    if p < 1.0 and random() >= p:
        return inputs

    else:
        return _flip_left_right(inputs)


def flip_up_down(inputs: dict, *, p: float = 1.0) -> dict:
    """Flip image up-down

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        p: probability of applying transformation

    """

    def _flip_up_down(inputs):
        H, W = _image_size(inputs["image"])

        inputs["image"] = inputs["image"][::-1, :, :]

        if "image_mask" in inputs:
            inputs["image_mask"] = inputs["image_mask"][::-1, :]

        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * [-1, 1] + [H, 0]

        #FIXME reformat bboxes so no negative length
        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * [-1, 1, -1, 1] + [H, 0, H, 0]

        # if "masks" in inputs:
        #     inputs["masks"] = inputs["masks"][:, ::-1, :, :]

        return inputs

    if p < 1.0 and random() >= p:
        return inputs

    else:
        return _flip_up_down(inputs)



from __future__ import annotations

from typing import Optional, Sequence, Callable

import flax.linen as nn
import jax
import jax.numpy as jnp
from flax.linen import normalization
from flax.linen import activation

from ..typing import DataDict, Array, ArrayLike
from .common import DropPath

""" Implements the convnext encoder. Described in https://arxiv.org/abs/2201.03545
Original implementation: https://github.com/facebookresearch/ConvNeXt
"""


class _Block(nn.Module):
    """ConvNeXt Block.
    Args:
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    drop_rate: int = 0.4
    layer_scale_init_value: float = 1e-6
    kernel_size: int = 7
    normalization: Callable[[None], nn.Module]=normalization.GroupNorm
    activation: Callable[[Array], Array]=activation.gelu

    @nn.compact
    def __call__(self, x: ArrayLike, *, training: Optional[bool] = None, **kwargs) -> Array:
        dim = x.shape[-1]
        ks = self.kernel_size
        scale = self.layer_scale_init_value

        shortcut = x

        x = nn.Conv(dim, (ks, ks), feature_group_count=dim)(x)
        x = self.normalization()(x)
        x = nn.Dense(dim * 4)(x)
        x = self.activation(x)
        x = nn.Dense(dim)(x)

        if scale > 0:
            gamma = self.param(
                "gamma", lambda rng, shape: scale * jnp.ones(shape), (x.shape[-1])
            )
            x = x * gamma

        deterministic = training is None or not training
        x = DropPath(self.drop_rate)(x, deterministic=deterministic)

        x = x + shortcut

        return x


# utility funcs

_imagenet_weights_urls = {
    "convnext_tiny_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_1k_224_ema.pth",
    "convnext_small_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_1k_224_ema.pth",
    "convnext_base_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_1k_224_ema.pth",
    "convnext_large_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_1k_224_ema.pth",
    "convnext_tiny_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_22k_224.pth",
    "convnext_small_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_22k_224.pth",
    "convnext_base_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_22k_224.pth",
    "convnext_large_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_22k_224.pth",
    "convnext_xlarge_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_xlarge_22k_224.pth",
}


class ConvNeXt(nn.Module):
    """ConvNeXt CNN backbone

    Attributes:
        patch_size: Stem patch size
        depths: Number of blocks at each stage.
        dims: Feature dimension at each stage.
        drop_path_rate: Stochastic depth rate.
        layer_scale_init_value: Init value for Layer Scale.
        out_channels:
            FPN output channels. Setting this to -1 disable the FPN, in which case
            the model output only encoder outputs.
    """

    patch_size: int = 4
    depths: Sequence[int] = (3, 3, 9, 3)
    dims: Sequence[int] = (96, 192, 384, 768)
    drop_path_rate: float = 0.4
    layer_scale_init_value: float = 1e-6
    kernel_size: int = 7
    normalization: Callable[[None], nn.Module]=nn.GroupNorm
    activation: Callable[[Array], Array]=nn.gelu

    @nn.compact
    def __call__(
        self, x: ArrayLike, *, training: Optional[bool] = None, **kwargs
    ) -> tuple[DataDict, DataDict]:
        """
        Args:
            x: Image input.
            training: Whether run the network in training mode (i.e. with stochastic depth)

        Returns:
            A list of feture arrays at different scaling, from higest to lowest resolution.
        """
        dp_rate = 0
        outputs = []
        for k in range(len(self.depths)):
            if k == 0:
                ps = self.patch_size
                x = nn.Conv(self.dims[k], (ps, ps), strides=(ps, ps))(x)
                x = self.normalization()(x)
            else:
                x = self.normalization()(x)
                x = nn.Conv(self.dims[k], (2, 2), strides=(2, 2))(x)

            for _ in range(self.depths[k]):
                x = _Block(
                    dp_rate,
                    layer_scale_init_value=self.layer_scale_init_value,
                    kernel_size=self.kernel_size,
                    normalization=self.normalization,
                    activation=self.activation,
                )(x, training=training)
                dp_rate += self.drop_path_rate / (sum(self.depths) - 1)

            outputs.append(x)

        return outputs

    @classmethod
    def get_model_tiny(cls, patch_size=4):
        return cls(patch_size, depths=(3, 3, 9, 3), dims=(96, 192, 384, 768))

    @classmethod
    def get_model_small(cls, patch_size=4):
        return cls(patch_size, depths=(3, 3, 27, 3), dims=(96, 192, 384, 768))

    @classmethod
    def get_model_base(cls, patch_size=4):
        return cls(patch_size, depths=(3, 3, 27, 3), dims=(128, 256, 512, 1024))

    @classmethod
    def get_model_large(cls, patch_size=4):
        return cls(patch_size, depths=(3, 3, 27, 3), dims=(192, 384, 768, 1536))

    @classmethod
    def get_model_x_large(cls, patch_size=4):
        return cls(patch_size, depths=(3, 3, 27, 3), dims=(256, 512, 1024, 2048))



""" 
Attributes:
    model_urls: URLs for build-in pretrain models. e.g model_urls["default"].
"""
from __future__ import annotations

import logging
import math
import time
from functools import partial, reduce
from typing import Mapping, Sequence, Tuple

import cv2
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np

from ..ops import patches_to_label, non_max_suppression, bboxes_of_patches, crop_and_resize_patches, box_iou_similarity
from ..utils import load_from_pretrained
from ..typing import Array, ArrayLike

Shape = Sequence[int]

model_urls: Mapping[str, str] = {
    "lacss2s-n-bf": "https://huggingface.co/jiyuuchc/lacss2s-n-bf/resolve/main/lacss2s-n-bf",
    "lacss2s-a": "https://huggingface.co/jiyuuchc/lacss2s-a/resolve/main/lacss2s-a",
    "lacss2s-b": "https://huggingface.co/jiyuuchc/lacss2s-b/resolve/main/lacss2s-b",
    "lacss2m-a": "https://huggingface.co/jiyuuchc/lacss2m-a/resolve/main/lacss2m-a",
}
model_urls["default"] = model_urls["lacss2s-b"]

def _remove_edge_instances(
    pred: dict,
    img_shape: Tuple[int, int],
    remove_top: bool = True,
    remove_bottom: bool = True,
    remove_left: bool = True,
    remove_right: bool = True,
    threshold: float = 0.5,
    min_border_pixel: int = 1,
):
    h, w = img_shape
    segs = pred["segmentations"] >= threshold
    yc, xc = jnp.mgrid[: segs.shape[-2], : segs.shape[-1]]
    yc += pred["y0s"][:, None, None]
    xc += pred["x0s"][:, None, None]

    removal = jnp.zeros([len(segs)], dtype=bool)

    removal |= remove_top & (
        jnp.count_nonzero((yc <= 0) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal |= remove_bottom & (
        jnp.count_nonzero((yc >= h - 1) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal |= remove_left & (
        jnp.count_nonzero((xc <= 0) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal |= remove_right & (
        jnp.count_nonzero((xc >= w - 1) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal = removal.reshape(pred["instance_mask"].shape)

    pred["instance_mask"] &= ~removal

    return pred


def _to_polygons(
    predictions, mask,
    *,
    scaling=1.0,
    segmentation_threshold=0,
    chain_approx=cv2.CHAIN_APPROX_SIMPLE,
) -> list:
    polygons = []

    y0s = np.asarray(predictions["segmentation_y0_coord"])
    x0s = np.asarray(predictions["segmentation_x0_coord"])
    segs = np.asarray(predictions["segmentations"] >= segmentation_threshold).astype("uint8")
    mask = np.asarray(mask)

    assert segs.shape[1] == 1
    segs = segs.squeeze(1)

    segs = segs[mask]
    y0s = y0s[mask]
    x0s = x0s[mask]

    for y0, x0, seg in zip(y0s, x0s, segs):
        c, _ = cv2.findContours(seg, cv2.RETR_EXTERNAL, chain_approx)

        max_len_element = reduce(
            lambda a, b: a if len(a) >= len(b) else b,
            c,
            np.zeros([0, 1, 2], dtype=int),
        )

        polygon = max_len_element.squeeze(1).astype(float) + [x0, y0]

        polygons.append(polygon)

    if scaling != 1.0:
        polygons = [(c + 0.5) / scaling - 0.5 for c in polygons]

    return polygons


def _draw_label(
    segs, y0s, x0s, img_shape, *,
    scaling=1.0,
    chain_approx=cv2.CHAIN_APPROX_SIMPLE,
):
    segs = segs.swapaxes(0,1)
    max_label = segs.shape[1]
    label = np.zeros(img_shape, dtype="uint16")

    for d, seg_slice in enumerate(segs):
        for k, (y0, x0, seg) in enumerate(zip(y0s[::-1], x0s[::-1], seg_slice[::-1])):
            c, _ = cv2.findContours(seg, cv2.RETR_EXTERNAL, chain_approx)

            if len(c) > 0:
                max_len_element = reduce(
                    lambda a, b: a if len(a) >= len(b) else b,
                    c,
                    np.zeros([0, 1, 2], dtype=int),
                )
                polygon = max_len_element.squeeze(1).astype(float) + [x0, y0]

                if scaling != 1.0:
                    polygon = (polygon + 0.5) / scaling - 0.5

                polygon = np.round(polygon).astype(int)
                cv2.fillPoly(label[d], [polygon], max_label - k)

    return label


def _compute_label(predictions, scaling, mask, image_shape, segmentation_threshold):
    if scaling == 1.0:
        return np.asarray(patches_to_label(
            predictions, 
            image_shape,
            mask=mask,
            score_threshold=0,
            threshold=segmentation_threshold,
        ))
    else:
        y0s = np.asarray(predictions["segmentation_y0_coord"])
        x0s = np.asarray(predictions["segmentation_x0_coord"])
        segs = np.asarray(predictions["segmentations"] >= segmentation_threshold).astype("uint8")
        segs, y0s, x0s = segs[mask], y0s[mask], x0s[mask]

        return _draw_label(
            segs, y0s, x0s, image_shape, scaling=scaling
        )


class Predictor:
    """Main class interface for model deployment. This is the only class you
    need if you don't train your own model

    Examples:
        The most common use case is to use a build-in pretrained model.

            import lacss.deploy

            # look up the url of a build-in mode
            url = lacss.deploy.model_urls["default"]

            # create the predictor instance
            predictor = lacss.deploy.Predictor(url)

            # make a prediction
            label = predictor.predict(image)

    Attributes:
        module: The underlying FLAX module
        params: Model weights.
    """
    DEFAULT_IMG_SIZE=544

    def __init__(
            self, 
            url: str | tuple[nn.Module, dict],
        ):
        """Construct Predictor

        Args:
            url: A URL or local path to the saved model.
                URLs for build-in pretrained models can be found in lacss.deploy.model_urls
        """

        if isinstance(url, tuple) and len(url) == 2:
            self.model = url
            if not isinstance(self.model[0], nn.Module):
                raise ValueError(
                    "Initiaize the Predictor with a tuple, but the first element is not a Module."
                )
        else:
            self.model = load_from_pretrained(url)

        self._apply_fn = jax.jit(partial(self.module.apply, dict(params=self.params)))

        logging.info(f"Prcompile the predictor for image shape {self.DEFAULT_IMG_SIZE, self.DEFAULT_IMG_SIZE}. This may take several minutes.")

        precompile_shape = [self.DEFAULT_IMG_SIZE] * 2 + [3]
        x = jnp.zeros(precompile_shape)
        _ = self.predict(x)


    def _format_image(self, image, scale=1.0):
        from skimage.transform import rescale

        if image.ndim == 3:
            image = image[None, ...]

        if scale != 1.0:
            image = rescale(image, [1.0, scale, scale], channel_axis=-1)

        image = image - image.mean()
        image = image / image.std()

        if image.shape[-1] == 1:
            image = np.repeat(image, 3, axis=-1)

        D, H, W, C = image.shape
        padto_h, padto_w = (H-1)//32*32+32, (W-1)//32*32+32
        if D == 1 and padto_h < self.DEFAULT_IMG_SIZE and padto_w < self.DEFAULT_IMG_SIZE:
            padto_h, padto_w = self.DEFAULT_IMG_SIZE, self.DEFAULT_IMG_SIZE
        padding = [[0, 0], [0, padto_h-H], [0, padto_w-W], [0, 3-C]]
        image = np.pad(image, padding)

        assert image.shape[-1] == 3, f"input image has more than 3 channels"

        depth = image.shape[0]
        if depth > 1:
            z_pad = 16 - depth if depth <= 16 else 32 - depth #FIXME avoid hardcoded padding
            image = np.pad(image, [[0, z_pad], [0,0], [0,0], [0,0]])

        mask_z = np.expand_dims(np.arange(image.shape[0]) < D, (1,2))
        mask_y = np.expand_dims(np.arange(image.shape[1]) < H, (0,2))
        mask_x = np.expand_dims(np.arange(image.shape[2]) < W, (0,1))
        input_mask = mask_z & mask_y & mask_x

        return image, input_mask


    def predict(
        self,
        image: ArrayLike,
        *,
        min_area: float = 0,
        scaling: float = 1.0,
        output_type: str = "label",
        score_threshold: float = 0.5,
        segmentation_threshold: float = 0.5,
        nms_iou: float = 1,
        remove_out_of_bound: bool|None = None,
    ) -> dict:
        """Predict segmentation.

        Args:
            image: A ndarray of (h,w,c) or (d,h,w,c) format. c must be 1-3

        Keyword Args:
            output_type: "label" | "contour" | "bbox"
            min_area: Minimum area of a valid prediction.
            scaling: A image scaling factor. If not 1, the input image will be resized internally before fed
                to the model. The results will be resized back to the scale of the orginal input image.
            score_threshold: Min score needed to be included in the output.
            segmentation_threshold: Threshold value for segmentation
            nms_iou: IOU threshold value for non-max-supression post-processing

        Returns:
            For "label" output:

                - pred_scores: The prediction scores of each instance.
                - pred_label: a 2D image label. 0 is background.

            For "contour" output:

                - pred_scores: The prediction scores of each instance.
                - pred_contours: a list of polygon arrays in x-y format.

            For "bbox" output (ie MaskRCNN):

                - pred_scores: The prediction scores of each instance.
                - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
                - pred_masks:  A 3d array representing (rescaled) segmentation mask within bboxes
        """
        logging.debug(f"started prediction with image {image.shape}")
        start_time = time.time()

        if remove_out_of_bound is not None:
            import warnings
            warnings.warn("remove_out_of_bound is deprecated", DeprecationWarning, 2)

        if image.ndim == 2:
            logging.warn("Input image has a ndim of 2. Reformate assuming this is a single-channel 2D image.")
            image = image[..., None]

        is_2d_input = image.ndim == 3 
        if is_2d_input:
            image = image[None, ...]

        orig_shape = image.shape

        image, input_mask = self._format_image(image, scaling)

        if not output_type in ("bbox", "label", "contour"):
            raise ValueError(
                f"output_type should be 'bbox'|'label'|'contour'. Got {output_type} instead."
            )
        
        if output_type == "contour" and image.shape[0] != 1:
            raise ValueError(f"contour output is only supported on 2D inputs")

        seg_logit_threshold = math.log(segmentation_threshold / (1 - segmentation_threshold))

        preds = self._apply_fn(image, mask=input_mask)["predictions"]
        instance_mask = np.asarray(preds["segmentation_is_valid"])
        instance_mask &= preds["scores"] >= score_threshold

        elapsed = (time.time() - start_time) * 1000
        logging.debug(f"finished model execution in {elapsed:.2f} ms")

        if min_area > 0:
            areas = jnp.count_nonzero(
                preds["segmentations"] > segmentation_threshold, axis=(1, 2, 3)
            )
            instance_mask &= areas >= min_area * scaling * scaling

        scores = preds["scores"]
        bboxes = bboxes_of_patches(preds, threshold=seg_logit_threshold) / scaling

        if nms_iou > 0 and nms_iou < 1:
            _, _, sel = non_max_suppression(
                scores, bboxes, bboxes.shape[0],
                threshold=nms_iou,
                return_selection=True,
                similarity_func=box_iou_similarity,
            )
            instance_mask = instance_mask & sel
        
        if output_type == "bbox":
            target_shape = (48, 48) if len(orig_shape) < 4 else (8, 48, 48)
            segmentations = crop_and_resize_patches(
                preds, target_shape, bboxes,
                convert_logits=True,
            )
            results = dict(
                pred_scores=np.asarray(scores)[instance_mask],
                pred_bboxes=np.asarray(bboxes)[instance_mask],
                pred_masks=np.asarray(segmentations)[instance_mask],
            )

        elif output_type == "contour":
            contours = _to_polygons(
                preds, instance_mask,
                scaling=scaling,
                segmentation_threshold=seg_logit_threshold,
            )
            results = dict(
                pred_scores=np.asarray(scores)[instance_mask],
                pred_bboxes=np.asarray(bboxes)[instance_mask],
                pred_contours=contours,
            )

        else:  # Label
            label = _compute_label(
                preds, scaling, instance_mask, 
                image.shape[:1] + orig_shape[-3:-1],
                seg_logit_threshold,
            )
            label = label[:orig_shape[0]]
            if is_2d_input:
                label = label.squeeze(0)
            results = dict(
                pred_scores=np.asarray(scores)[instance_mask],
                pred_label=label,
            )

        elapsed = (time.time() - start_time) * 1000
        logging.debug(f"finished prediction in {elapsed:.2f} ms")
        
        return results

    # def predict_on_large_image(
    #     self,
    #     image: ArrayLike,
    #     gs: int,
    #     ss: int,
    #     *,
    #     scaling: float = 1,
    #     min_area: int = 0,
    #     nms_iou: float = 0.25,
    #     segmentation_threshold: float = 0.5,
    #     score_threshold: float = 0.5,
    #     output_type: str = "label",
    #     min_cells_per_patch: int = 0,
    #     disable_padding: bool = False,
    #     **kwargs,
    # ) -> dict:
    #     """Make prediction on very large image by dividing into a grid.

    #     Direct model prediction on large image may cause out-of-memory error. This
    #     method divided the large image to smaller grid and then stitch the results
    #     to form the complete prediction.

    #     Args:
    #         image: An image with (H, W, C) format.
    #         gs: An int value. Grid size of the computation.
    #         ss: An int value of stepping size. Must be small than gs to produce valid
    #             results.

    #     Keyword Args:
    #         output_type: label" | "contour" | "bbox"
    #         min_area: Minimum area of a valid prediction.
    #         scaling: A image scaling factor. If not 1, the input image will be resized internally before fed
    #             to the model. The results will be resized back to the scale of the orginal input image.
    #         score_threshold: Min score needed to be included in the output.
    #         segmentation_threshold: Threshold value for segmentation.
    #         nms_iou: IOU threshold value for non-max-supression durint post-processing

    #     Returns:
    #         For "label" output:

    #             - pred_scores: The prediction scores of each instance.
    #             - pred_label: a 2D image label. 0 is background.

    #         For "contour" output:

    #             - pred_scores: The prediction scores of each instance.
    #             - pred_contours: a list of polygon arrays in x-y format.

    #         For "bbox" output (ie MaskRCNN):

    #             - pred_scores: The prediction scores of each instance.
    #             - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
    #             - pred_masks: A 3d array representing segmentation within bboxes
    #     """

    #     if not output_type in ("bbox", "label", "contour"):
    #         raise ValueError(
    #             f"output_type should be 'patch'|'label'|'contour'. Got {output_type} instead."
    #         )

    #     module, params = self.model

    #     if len(kwargs) > 0:
    #         apply_fn = _cached_partial(module.apply, **kwargs)
    #     else:
    #         apply_fn = module.apply

    #     inv_sig_st = math.log(segmentation_threshold / (1 - segmentation_threshold))
    #     ctx = freeze(
    #         dict(
    #             apply_fn=apply_fn,
    #             output_type="bbox" if output_type == "bbox" else "grid",
    #             segmentation_threshold=inv_sig_st,
    #             score_threshold=score_threshold,
    #             scaling=scaling,
    #             min_area=min_area,
    #             remove_out_of_bound=True,
    #             nms_iou=0,
    #         )
    #     )

    #     h, w = image.shape[:2]

    #     if not disable_padding:
    #         get_padding = lambda a, gs, ss: (a - 1) // ss * ss + gs - a

    #         padded_img = np.pad(
    #             image,
    #             [[0, get_padding(h, gs, ss)], [0, get_padding(w, gs, ss)], [0, 0]],
    #         )
    #     else:
    #         padded_img = image

    #     preds = []
    #     for y0 in range(0, h, ss):
    #         for x0 in range(0, w, ss):
    #             logging.info(f"Processing grid {y0}-{x0}")
    #             pred = _predict(
    #                 ctx,
    #                 params,
    #                 padded_img[y0 : y0 + gs, x0 : x0 + gs],
    #             )
    #             pred = _remove_edge_instances(
    #                 pred,
    #                 [gs, gs],
    #                 remove_top=y0 > 0,
    #                 remove_bottom=(y0 + gs) < h,
    #                 remove_left=x0 > 0,
    #                 remove_right=(x0 + gs) < w,
    #                 threshold=segmentation_threshold,
    #             )

    #             logging.info(f"Transfer result...")

    #             mask = np.asarray(pred["instance_mask"])

    #             if np.count_nonzero(mask) > min_cells_per_patch:
    #                 if output_type != "bbox":
    #                     pred = dict(
    #                         scores=np.array(pred["scores"])[mask],
    #                         bboxes=np.array(pred["bboxes"])[mask] + [y0, x0, y0, x0],
    #                         segmentations=np.array(pred["segmentations"])[mask],
    #                         y0s=np.array(pred["y0s"])[mask] + y0,
    #                         x0s=np.array(pred["x0s"])[mask] + x0,
    #                     )

    #                     polygons = _to_polygons(
    #                         pred,
    #                         segmentation_threshold=segmentation_threshold,
    #                         scaling=scaling,
    #                     )

    #                     valid_polygons = np.asarray([len(p) > 0 for p in polygons])
    #                     scores = np.where(valid_polygons, pred["scores"], -1)
    #                     bboxes = np.where(
    #                         valid_polygons[:, None], pred["bboxes"] / scaling, -1
    #                     )

    #                     preds.append(
    #                         dict(
    #                             scores=scores,
    #                             polygons=np.fromiter(polygons, dtype=object),
    #                             bboxes=bboxes,
    #                         )
    #                     )

    #                 else:
    #                     bboxes = np.array(pred["bboxes"])[mask] + [y0, x0, y0, x0]
    #                     bboxes /= scaling
    #                     preds.append(
    #                         dict(
    #                             scores=np.array(pred["scores"])[mask],
    #                             bboxes=bboxes,
    #                             segmentations=np.array(pred["segmentations"])[mask],
    #                         )
    #                     )

    #     logging.info(f"Postprocessing...")
    #     preds = jax.tree_util.tree_map(lambda *x: np.concatenate(x), *preds)

    #     # sort based on scores
    #     asort = np.argsort(preds["scores"])[::-1]
    #     preds = jax.tree_util.tree_map(lambda x: x[asort], preds)

    #     # nms
    #     if nms_iou > 0:
    #         logging.info(f"nms...")
    #         scores = preds["scores"]
    #         boxes = preds["bboxes"]
    #         _, _, selections = non_max_suppression(
    #             scores,
    #             boxes,
    #             -1,
    #             threshold=nms_iou,
    #             min_score=0,
    #             return_selection=True,
    #         )

    #         preds = jax.tree_util.tree_map(lambda x: x[selections], preds)

    #     if output_type == "bbox":  # FIXME unimplemented

    #         return dict(
    #             pred_scores=preds["scores"],
    #             pred_bboxes=preds["bboxes"],
    #             pred_masks=segs,
    #         )

    #     if output_type == "label":
    #         logging.info(f"Generating label...")

    #         label = _draw_label(preds["polygons"], image.shape[:2])

    #         return dict(
    #             pred_scores=preds["scores"],
    #             pred_label=label,
    #         )

    #     else:  # contours
    #         return dict(
    #             pred_scores=preds["scores"],
    #             pred_contours=preds["polygons"],
    #         )

    @property
    def module(self) -> nn.Module:
        return self.model[0]

    @property
    def params(self) -> dict:
        return self.model[1]

    @params.setter
    def params(self, new_params):
        self.model = self.module, new_params



""" Ops on bounding-boxes

All functions here are degisned to work as either a numpy op or a jax op depending on
the data type of the input.
"""
from __future__ import annotations

import jax.numpy as jnp
import numpy as np

from ..typing import *


def box_area(box: ArrayLike) -> ArrayLike:
    """Computes area of boxes.
    Args:
      box: a float Tensor with [..., N, 2d].

    Returns:
      a float Tensor with [..., N]
    """
    ndim = box.shape[-1] // 2
    assert ndim * 2 == box.shape[-1]

    min_vals = box[..., :ndim]
    max_vals = box[..., ndim:]

    return (max_vals - min_vals).prod(axis=-1)


def box_intersection(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """Compute pairwise intersection areas between boxes.

    Args:
      gt_boxes: [..., N, 2d]
      boxes: [..., M, 2d]

    Returns:
      a float Tensor with shape [..., N, M] representing pairwise intersections.
    """
    if isinstance(gt_boxes, jnp.ndarray) & isinstance(boxes, jnp.ndarray):
        minimum = jnp.minimum
        maximum = jnp.maximum
    else:
        minimum = np.minimum
        maximum = np.maximum

    ndim = gt_boxes.shape[-1] // 2
    assert ndim * 2 == gt_boxes.shape[-1]
    assert ndim * 2 == boxes.shape[-1]

    min_vals_1 = gt_boxes[..., None, :ndim] # [..., N, 1, d]
    max_vals_1 = gt_boxes[..., None, ndim:]
    min_vals_2 = boxes[..., None, :, :ndim] # [..., 1, M, d]
    max_vals_2 = boxes[..., None, :, ndim:]

    min_max = minimum(max_vals_1, max_vals_2) #[..., N, M, d]
    max_min = maximum(min_vals_1, min_vals_2)

    intersects = maximum(0, min_max - max_min) # [..., N, M, d]

    return intersects.prod(axis=-1)


def box_iou_similarity(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """Computes pairwise intersection-over-union between box collections.

    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., M, 2d].

    Returns:
      a Tensor with shape [..., N, M] representing pairwise iou scores.
    """
    intersections = box_intersection(gt_boxes, boxes)
    gt_boxes_areas = box_area(gt_boxes)
    boxes_areas = box_area(boxes)
    unions = gt_boxes_areas[..., None] + boxes_areas[..., None, :] - intersections

    ious = intersections / (unions + 1e-6)

    return ious

def iou_loss(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """ IOU loss = 1 - IOU
    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., M, 2d].

    Returns:
      a Tensor with shape [..., N, M] representing pairwise loss.
    """
    return 1 - box_iou_similarity(gt_boxes, boxes)


def generalized_iou_loss(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """ Loss_GIoU = 1 - IoU + |C - B union B_GT| / |C|
    where C is the smallest enclosing box for both B and B_GT. The resulting value has a gradient 
    for non-overlapping boxes. See  Zheng et al. [AAAI 2020] 

    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., M, 2d].

    Returns:
      a Tensor with shape [..., N, M] representing pairwise loss.
    """
    if isinstance(gt_boxes, jnp.ndarray) & isinstance(boxes, jnp.ndarray):
        minimum = jnp.minimum
        maximum = jnp.maximum
    else:
        minimum = np.minimum
        maximum = np.maximum

    ndim = gt_boxes.shape[-1] // 2
    assert ndim * 2 == gt_boxes.shape[-1]
    assert ndim * 2 == boxes.shape[-1]

    intersections = box_intersection(gt_boxes, boxes)
    gt_boxes_areas = box_area(gt_boxes)
    boxes_areas = box_area(boxes)
    unions = gt_boxes_areas[..., None] + boxes_areas[..., None, :] - intersections

    ious = intersections / (unions + 1e-6)

    mins = minimum(gt_boxes[..., None, :ndim], boxes[..., None, :, :ndim])
    maxs = maximum(gt_boxes[..., None, ndim:], boxes[..., None, :, ndim:])
    intersects = maximum(0, maxs - mins) # [..., N, M, d]

    C = intersects.prod(axis=-1)
    
    return 1 - ious + (C - unions) / (C + 1e-6)

def distance_iou_loss(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """ Loss_distance_iou = 1 - IoU + \rho2(B, B_GT) / c2
    The correction term is "distance_sq between box center" / "distance_sq of enclosing box cornor"

    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., M, 2d].

    Returns:
      a Tensor with shape [..., N, M] representing pairwise loss.    
    """
    if isinstance(gt_boxes, jnp.ndarray) & isinstance(boxes, jnp.ndarray):
        minimum = jnp.minimum
        maximum = jnp.maximum
    else:
        minimum = np.minimum
        maximum = np.maximum

    ndim = gt_boxes.shape[-1] // 2
    assert ndim * 2 == gt_boxes.shape[-1]
    assert ndim * 2 == boxes.shape[-1]

    ious = box_iou_similarity(gt_boxes, boxes)
    mins = minimum(gt_boxes[..., None, :ndim], boxes[..., None, :, :ndim])
    maxs = maximum(gt_boxes[..., None, ndim:], boxes[..., None, :, ndim:])
    intersects = maximum(0, maxs - mins) # [..., N, M, d]
    c2 = (intersects ** 2).sum(axis=-1)

    center_gt = (gt_boxes[..., :ndim] + gt_boxes[..., ndim:]) / 2
    center = (boxes[..., :ndim] + boxes[..., ndim:]) / 2
    rho2 = ((center_gt[..., None, :] - center[..., None, :, :]) ** 2).sum(axis=-1)

    return 1 - ious + rho2/c2



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class DType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FLOAT32: _ClassVar[DType]
FLOAT32: DType

class Image(_message.Message):
    __slots__ = ("height", "width", "channel", "dtype", "data")
    HEIGHT_FIELD_NUMBER: _ClassVar[int]
    WIDTH_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    height: int
    width: int
    channel: int
    dtype: DType
    data: bytes
    def __init__(self, height: _Optional[int] = ..., width: _Optional[int] = ..., channel: _Optional[int] = ..., dtype: _Optional[_Union[DType, str]] = ..., data: _Optional[bytes] = ...) -> None: ...

class Settings(_message.Message):
    __slots__ = ("min_cell_area", "remove_out_of_bound", "scaling", "nms_iou", "detection_threshold", "segmentation_threshold", "return_polygon")
    MIN_CELL_AREA_FIELD_NUMBER: _ClassVar[int]
    REMOVE_OUT_OF_BOUND_FIELD_NUMBER: _ClassVar[int]
    SCALING_FIELD_NUMBER: _ClassVar[int]
    NMS_IOU_FIELD_NUMBER: _ClassVar[int]
    DETECTION_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    SEGMENTATION_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    RETURN_POLYGON_FIELD_NUMBER: _ClassVar[int]
    min_cell_area: float
    remove_out_of_bound: bool
    scaling: float
    nms_iou: float
    detection_threshold: float
    segmentation_threshold: float
    return_polygon: bool
    def __init__(self, min_cell_area: _Optional[float] = ..., remove_out_of_bound: bool = ..., scaling: _Optional[float] = ..., nms_iou: _Optional[float] = ..., detection_threshold: _Optional[float] = ..., segmentation_threshold: _Optional[float] = ..., return_polygon: bool = ...) -> None: ...

class Input(_message.Message):
    __slots__ = ("settings", "image")
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    IMAGE_FIELD_NUMBER: _ClassVar[int]
    settings: Settings
    image: Image
    def __init__(self, settings: _Optional[_Union[Settings, _Mapping]] = ..., image: _Optional[_Union[Image, _Mapping]] = ...) -> None: ...

class Label(_message.Message):
    __slots__ = ("height", "width", "data")
    HEIGHT_FIELD_NUMBER: _ClassVar[int]
    WIDTH_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    height: int
    width: int
    data: bytes
    def __init__(self, height: _Optional[int] = ..., width: _Optional[int] = ..., data: _Optional[bytes] = ...) -> None: ...

class Result(_message.Message):
    __slots__ = ("score", "label")
    SCORE_FIELD_NUMBER: _ClassVar[int]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    score: Label
    label: Label
    def __init__(self, score: _Optional[_Union[Label, _Mapping]] = ..., label: _Optional[_Union[Label, _Mapping]] = ...) -> None: ...

class Point(_message.Message):
    __slots__ = ("x", "y")
    X_FIELD_NUMBER: _ClassVar[int]
    Y_FIELD_NUMBER: _ClassVar[int]
    x: float
    y: float
    def __init__(self, x: _Optional[float] = ..., y: _Optional[float] = ...) -> None: ...

class Polygon(_message.Message):
    __slots__ = ("score", "points")
    SCORE_FIELD_NUMBER: _ClassVar[int]
    POINTS_FIELD_NUMBER: _ClassVar[int]
    score: float
    points: _containers.RepeatedCompositeFieldContainer[Point]
    def __init__(self, score: _Optional[float] = ..., points: _Optional[_Iterable[_Union[Point, _Mapping]]] = ...) -> None: ...

class PolygonResult(_message.Message):
    __slots__ = ("polygons",)
    POLYGONS_FIELD_NUMBER: _ClassVar[int]
    polygons: _containers.RepeatedCompositeFieldContainer[Polygon]
    def __init__(self, polygons: _Optional[_Iterable[_Union[Polygon, _Mapping]]] = ...) -> None: ...



import logging
from concurrent import futures
from pathlib import Path

import grpc
import jax
import numpy as np
import typer

from .proto import lacss_pb2, lacss_pb2_grpc

_AUTH_HEADER_KEY = "authorization"

app = typer.Typer(pretty_exceptions_enable=False)


def process_input(msg):
    image = msg.image
    np_img = np.frombuffer(image.data, dtype=">f4").astype("float32")
    if image.depth == 0 or image.depth == 1:
        np_img = np_img.reshape(image.height, image.width, image.channel)
    else:
        np_img = np_img.reshape(image.depth, image.height, image.width, image.channel)

    return np_img, msg.settings


def process_result(preds, is3d):
    msg = lacss_pb2.PolygonResult()

    scores = preds["pred_scores"]
    if is3d:
        from skimage.measure import regionprops
        label = preds["pred_label"].astype(int)

        for rp, score in zip(regionprops(label), scores):
            polygon_msg = lacss_pb2.Polygon()
            polygon_msg.score = score
            point = lacss_pb2.Point()
            point.z, point.y, point.x = rp.centroid
            polygon_msg.points.append(point)
    
            msg.polygons.append(polygon_msg)

    else:
        polygons = preds["pred_contours"]
        for polygon, score in zip(polygons, scores):
            if len(polygon) > 0:
                polygon_msg = lacss_pb2.Polygon()
                polygon_msg.score = score
                for p in polygon:
                    point = lacss_pb2.Point()
                    point.x = p[0]
                    point.y = p[1]
                    polygon_msg.points.append(point)
                msg.polygons.append(polygon_msg)

    return msg


class TokenValidationInterceptor(grpc.ServerInterceptor):
    def __init__(self, token):
        def abort(ignored_request, context):
            context.abort(grpc.StatusCode.UNAUTHENTICATED, "Invalid token signature")

        self._abort_handler = grpc.unary_unary_rpc_method_handler(abort)
        self.token = token

    def intercept_service(self, continuation, handler_call_details):
        expected_metadata = (_AUTH_HEADER_KEY, f"Bearer {self.token}")
        if self.token is None or expected_metadata in handler_call_details.invocation_metadata:
            return continuation(handler_call_details)
        else:
            return self._abort_handler

class LacssServicer(lacss_pb2_grpc.LacssServicer):
    def __init__(self, model):
        self.model = model

    def RunDetection(self, request, context):
        img, settings = process_input(request)
        is3d = img.ndim == 4
        logging.debug(f"received image {img.shape}")

        preds = self.model.predict(
            img,
            min_area=settings.min_cell_area,
            scaling=settings.scaling,
            nms_iou=settings.nms_iou,
            score_threshold=settings.detection_threshold,
            segmentation_threshold=settings.segmentation_threshold,
            output_type="label" if is3d else "contour" ,
        )

        result = process_result(preds, is3d)

        logging.debug(f"Reply with message of size {result.ByteSize()}")

        return result


def get_predictor(modelpath):
    from .predict import Predictor

    model = Predictor(modelpath)

    logging.info(f"lacss_server: loaded model from {modelpath}")

    model.module.detector.max_output = 512  # FIXME good default?
    model.module.detector.min_score = 0.2

    return model

def show_urls():
    from .predict import model_urls
    
    print("Pretrained model files:")
    print("==============================")
    for k, v in model_urls.items():
        print(f"{k}: {v}")
    print()


@app.command()
def main(
    modelpath: Path|None = None,
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool|None = None,
    debug: bool = False,
    compression: bool = True,
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    if modelpath is None:
        show_urls()
        return

    model = get_predictor(modelpath)

    logging.info(f"lacss_server: default backend is {jax.default_backend()}")

    if jax.default_backend() == "cpu":
        logging.warn(
            f"lacss_server: WARNING: No GPU configuration. This might be very slow ..."
        )

    if token is None:
        token = not local
    if token:
        import secrets

        token = secrets.token_urlsafe(32)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token}")
        print("=======================================================================")
        print()
    else:
        token = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip if compression else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token),),
        options=(("grpc.max_receive_message_length", 1024*1024*16),),
    )

    lacss_pb2_grpc.add_LacssServicer_to_server(LacssServicer(model), server)

    if local:
        server.add_secure_port(f"localhost:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logging.info(f"lacss_server: listening on port {port}")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    app()
